"""
To run this task:

seb run -t MuniIntent -m multilingual-e5-large-instruct,multilingual-e5-large,paraphrase-multilingual-mpnet-base-v2 -c src/experimental_tasks/muni_intent_classification.py
seb run -t MuniIntent -m paraphrase-multilingual-mpnet-base-v2 -c src/experimental_tasks/muni_intent_classification.py
"""

from typing import Any

from mteb.abstasks import AbsTaskClassification
from seb.interfaces.mteb_task import MTEBTask
from seb.interfaces.task import Task
from seb.registries import tasks


class MuniIntentClassification(AbsTaskClassification):
    @property
    def description(self) -> dict[str, Any]:
        return {
            "name": "MuniIntentClassification",
            "hf_hub_name": "KennethEnevoldsen/muni_intents",
            "description": "An intent classification dataset used for the MUNI chatbot.",
            "reference": "NA",
            "type": "Classification",
            "category": "s2s",
            "eval_splits": ["test"],
            "eval_langs": ["da"],  # assumed to be bokmål
            "main_score": "accuracy",
            "n_experiments": 10,
            "samples_per_label": 16,
            "revision": "f342548ed8ed4e1b3a31c546a34d2f6fd1a6ed12",
        }


@tasks.register("MuniIntent")
def create_swedn_sts() -> Task:
    task = MTEBTask(MuniIntentClassification())
    task.name = "MuniIntent"
    task.version = "0.0.1"
    task.domain = ["non-fiction"]
    return task



from __future__ import annotations
import logging
from datetime import date
from functools import partial
import torch
from typing import Any, Optional, TypeVar, Union, List
from collections.abc import Iterable, Sequence
from tqdm import tqdm
from itertools import islice
import numpy as np

import seb
from seb.interfaces.model import LazyLoadEncoder, ModelMeta, SebModel, Encoder
from seb.interfaces.task import Task
from seb.registries import models


logger = logging.getLogger(__name__)
T = TypeVar("T")


def task_to_instruction(task: Task) -> str:
    if task.task_type in ["STS"]:
        return "Retrieve semantically similar text"
    if task.task_type in ["Summarization"]:
        return "Given a news summary, retrieve other semantically similar summaries"
    if task.task_type in ["BitextMining"]:
        task_name_to_instruct: dict[str, str] = {
            "Bornholm Parallel": "Retrieve parallel sentences in Danish and Bornholmsk",
            "Norwegian courts": "Retrieve parallel sentences in Norwegian Bokmål and Nynorsk",
        }
        default_instruction = "Retrieve parallel sentences."
        return task_name_to_instruct.get(task.name, default_instruction)
    if task.task_type in ["Classification"]:
        task_name_to_instruct: dict[str, str] = {
            "Angry Tweets": "Classify Danish tweets by sentiment. (positive, negative, neutral)",
            "DKHate": "Classify Danish tweets based on offensiveness (offensive, not offensive)",
            "Da Political Comments": "Classify Danish political comments for sentiment",
            "DaLAJ": "Classify texts based on linguistic acceptability in Swedish",
            "LCC": "Classify texts based on sentiment",
            "Language Identification": "Classify texts based on language",
            "Massive Intent": "Given a user utterance as query, find the user intents",
            "Massive Scenario": "Given a user utterance as query, find the user scenarios",
            "NoReC": "Classify Norwegian reviews by sentiment",
            "SweReC": "Classify Swedish reviews by sentiment",
            "Norwegian parliament": "Classify parliament speeches in Norwegian based on political affiliation",
            "ScaLA": "Classify passages in Scandinavian Languages based on linguistic acceptability",
        }
        default_instruction = "Classify user passages"
        return task_name_to_instruct.get(task.name, default_instruction)
    if task.task_type in ["Clustering"]:
        task_name_to_instruct: dict[str, str] = {
            "ArxivClusteringP2P": "Identify the main and secondary category of Arxiv papers based on the titles and abstracts",
            "VG Clustering": "Identify the categories (e.g. sports) of given articles in Norwegian",
            "SNL Clustering": "Identify categories in a Norwegian lexicon",
            "SwednClustering": "Identify news categories in Swedish passages",
        }
        default_instruction = "Identify categories in user passages"
        return task_name_to_instruct.get(task.name, default_instruction)
    if task.task_type in ["Reranking"]:
        return "Retrieve semantically similar passages."
    if task.task_type in ["Retrieval"]:
        task_name_to_instruct: dict[str, str] = {
            "Twitterhjerne": "Retrieve answers to questions asked in Danish tweets",
            "SwednRetrieval": "Given a Swedish news headline retrieve summaries or news articles",
            "TV2Nord Retrieval": "Given a summary of a Danish news article retrieve the corresponding news article",
            "DanFEVER": "Given a claim in Danish, retrieve documents that support the claim",
            "SNL Retrieval": "Given a lexicon headline in Norwegian, retrieve its article",
            "NorQuad": "Given a question in Norwegian, retrieve the answer from Wikipedia articles",
            "SweFAQ": "Retrieve answers given questions in Swedish",
            "ArguAna": "Given a claim, find documents that refute the claim",
            "ClimateFEVER": "Given a claim about climate change, retrieve documents that support or refute the claim",
        }
        default_instruction = "Retrieve text based on user query."
        return task_name_to_instruct.get(task.name, default_instruction)
    return ""


def batched(iterable: Iterable[T], n: int) -> Iterable[tuple[T, ...]]:
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


class LLM2VecModel(Encoder):
    def __init__(
        self,
        base_model_name_or_path: str,
        peft_model_name_or_path: str,
        max_length: int,
        max_batch_size: Optional[int] = None,
        **kwargs: Any,
    ):
        logger.info("Started loading LLM2Vec model")
        try:
            from llm2vec import LLM2Vec
        except ImportError:
            raise ImportError("To use the LLM2Vec models `llm2vec` is required. Please install it with `pip seb[llm2vec].")
        extra_kwargs = {}
        try:
            import flash_attn  # noqa

            extra_kwargs["attn_implementation"] = "flash_attention_2"
        except ImportError:
            logger.warning(
                "LLM2Vec models were trained with flash attention enabled. For optimal performance, please install the `flash_attn` package"
            )
        self.model = LLM2Vec.from_pretrained(
            base_model_name_or_path=base_model_name_or_path,
            peft_model_name_or_path=peft_model_name_or_path,
            **extra_kwargs,
            **kwargs,
            device_map="auto",
            torch_dtype=torch.bfloat16,
        )
        self.max_length = max_length
        self.max_batch_size = max_batch_size

    def preprocess(self, sentences, instruction) -> list[str] | list[list[str]]:
        if instruction is not None:
            sentences = [[instruction, sentence] for sentence in sentences]
        return sentences

    def encode(
        self,
        sentences: list[str],
        *,
        task: Optional[Task] = None,
        batch_size: int = 32,
        **kwargs: Any,
    ) -> np.ndarray:
        if self.max_batch_size and batch_size > self.max_batch_size:
            batch_size = self.max_batch_size

        batched_embeddings = []
        if task is not None:
            instruction = task_to_instruction(task)
        else:
            instruction = None

        for batch in tqdm(batched(sentences, batch_size)):
            preprocessed_batch = self.preprocess(batch, instruction=instruction)
            with torch.inference_mode():
                embedded_batch = self.model.encode(preprocessed_batch)
            batched_embeddings.append(embedded_batch)

        return torch.cat(batched_embeddings).numpy()


@models.register("TTC-L2V-supervised-da-1")
def create_llm2vec_da_mntp_ttc_supervised() -> SebModel:
    base_model = "jealk/llm2vec-da-mntp"
    peft_model = "jealk/TTC-L2V-supervised-1"
    meta = ModelMeta(
        name="TTC-L2V-supervised-da-1",
        huggingface_name=peft_model,
        reference=f"https://huggingface.co/{peft_model}",
        languages=["da"],
        open_source=True,
        embedding_size=4096,
        architecture="LLM2Vec",
        release_date=date(2024, 12, 20),
    )
    partial_model = partial(
        LLM2VecModel,
        base_model_name_or_path=base_model,
        peft_model_name_or_path=peft_model,
        max_length=8192,
    )
    return SebModel(
        encoder=LazyLoadEncoder(partial_model),
        meta=meta,
    )


@models.register("TTC-L2V-unsupervised-da-1")
def create_llm2vec_da_mntp_ttc_unsupervised() -> SebModel:
    base_model = "jealk/llm2vec-da-mntp"
    peft_model = "jealk/TTC-L2V-unsupervised-1"
    meta = ModelMeta(
        name="TTC-L2V-unsupervised-da-1",
        huggingface_name=peft_model,
        reference=f"https://huggingface.co/{peft_model}",
        languages=["da"],
        open_source=True,
        embedding_size=4096,
        architecture="LLM2Vec",
        release_date=date(2024, 12, 20),
    )
    partial_model = partial(
        LLM2VecModel,
        base_model_name_or_path=base_model,
        peft_model_name_or_path=peft_model,
        max_length=8192,
    )
    return SebModel(
        encoder=LazyLoadEncoder(partial_model),
        meta=meta,
    )


@models.register("TTC-L2V-supervised-2")
def create_llm2vec_da_mntp_ttc_supervised() -> SebModel:
    base_model = "jealk/llm2vec-da-mntp-2"
    peft_model = "jealk/TTC-L2V-supervised-2"
    meta = ModelMeta(
        name="TTC-L2V-supervised-2",
        huggingface_name=peft_model,
        reference=f"https://huggingface.co/{peft_model}",
        languages=["da", "sv", "nb"],
        open_source=True,
        embedding_size=4096,
        architecture="LLM2Vec",
        release_date=date(2025, 5, 16),
    )
    partial_model = partial(
        LLM2VecModel,
        base_model_name_or_path=base_model,
        peft_model_name_or_path=peft_model,
        max_length=8192,
    )
    return SebModel(
        encoder=LazyLoadEncoder(partial_model),
        meta=meta,
    )



from .cohere_models import *
from .e5_instruct_models import *
from .e5_models import *
from .fairseq_models import *
from .fasttext_models import *
from .sentence_transformer_models import *
from .openai_models import *
from .translate_e5_models import *
from .voyage_models import *
from .bge_models import *
from .jina_models import *
from .arctic_models import *
from .llm2vec_models import *



import logging
from collections.abc import Iterable, Sequence
from datetime import date
from functools import partial
from itertools import islice
from typing import Any, Literal, Optional, TypeVar

import numpy as np
import torch
from torch import Tensor
from tqdm import tqdm
from transformers import AutoModel, AutoTokenizer, BatchEncoding
from transformers.modeling_outputs import ModelOutput

from seb.interfaces.model import Encoder, LazyLoadEncoder, ModelMeta, SebModel
from seb.interfaces.task import Task
from seb.registries import models

logger = logging.getLogger(__name__)


T = TypeVar("T")
EncodeTypes = Literal["query", "passage"]


def batched(iterable: Iterable[T], n: int) -> Iterable[tuple[T, ...]]:
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


def batch_to_device(batch_data: dict[str, torch.Tensor], device: str = "cuda") -> dict[str, torch.Tensor]:
    return {key: data.to(device) for key, data in batch_data.items()}


def task_to_instruction(task: Task) -> str:
    if task.task_type in ["STS"]:
        return "Retrieve semantically similar text"
    if task.task_type in ["Summarization"]:
        return "Given a news summary, retrieve other semantically similar summaries"
    if task.task_type in ["BitextMining"]:
        task_name_to_instruct: dict[str, str] = {
            "Bornholm Parallel": "Retrieve parallel sentences in Danish and Bornholmsk",
            "Norwegian courts": "Retrieve parallel sentences in Norwegian Bokmål and Nynorsk",
        }
        default_instruction = "Retrieve parallel sentences."
        return task_name_to_instruct.get(task.name, default_instruction)
    if task.task_type in ["Classification"]:
        task_name_to_instruct: dict[str, str] = {
            "Angry Tweets": "Classify Danish tweets by sentiment. (positive, negative, neutral)",
            "DKHate": "Classify Danish tweets based on offensiveness (offensive, not offensive)",
            "Da Political Comments": "Classify Danish political comments for sentiment",
            "DaLAJ": "Classify texts based on linguistic acceptability in Swedish",
            "LCC": "Classify texts based on sentiment",
            "Language Identification": "Classify texts based on language",
            "Massive Intent": "Given a user utterance as query, find the user intents",
            "Massive Scenario": "Given a user utterance as query, find the user scenarios",
            "NoReC": "Classify Norwegian reviews by sentiment",
            "SweReC": "Classify Swedish reviews by sentiment",
            "Norwegian parliament": "Classify parliament speeches in Norwegian based on political affiliation",
            "ScaLA": "Classify passages in Scandinavian Languages based on linguistic acceptability",
        }
        default_instruction = "Classify user passages"
        return task_name_to_instruct.get(task.name, default_instruction)
    if task.task_type in ["Clustering"]:
        task_name_to_instruct: dict[str, str] = {
            "ArxivClusteringP2P": "Identify the main and secondary category of Arxiv papers based on the titles and abstracts",
            "VG Clustering": "Identify the categories (e.g. sports) of given articles in Norwegian",
            "SwednClustering": "Identify news categories in Swedish passages",
            "HistoricalDanishClustering": "Identify the author of a given passage from historical Danish fiction.",
        }
        default_instruction = "Identify categories in user passages"
        return task_name_to_instruct.get(task.name, default_instruction)
    if task.task_type in ["Reranking"]:
        return "Retrieve semantically similar passages."
    if task.task_type in ["Retrieval"]:
        task_name_to_instruct: dict[str, str] = {
            "Twitterhjerne": "Retrieve answers to questions asked in Danish tweets",
            "SwednRetrieval": "Given a Swedish news headline retrieve summaries or news articles",
            "TV2Nord Retrieval": "Given a summary of a Danish news article retrieve the corresponding news article",
            "DanFEVER": "Given a claim in Danish, retrieve documents that support the claim",
            "NorQuad": "Given a question in Norwegian, retrieve the answer from Wikipedia articles",
            "SweFAQ": "Retrieve answers given questions in Swedish",
            "ArguAna": "Given a claim, find documents that refute the claim",
            "ClimateFEVER": "Given a claim about climate change, retrieve documents that support or refute the claim",
        }
        default_instruction = "Retrieve text based on user query."
        return task_name_to_instruct.get(task.name, default_instruction)
    return ""


class E5Instruct(Encoder):
    def __init__(self, model_name: str, max_length: int, max_batch_size: Optional[int] = None, **kwargs: Any):
        logger.info("Started loading e5 instruct model")
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModel.from_pretrained(model_name, **kwargs)
        self.max_length = max_length
        self.max_batch_size = max_batch_size

    def preprocess(self, sentences: Sequence[str], instruction: str, encode_type: EncodeTypes) -> BatchEncoding:
        if encode_type == "query":
            sentences = [f"Instruction: {instruction}\nQuery: {sentence}" for sentence in sentences]

        batch_dict = self.tokenizer(
            sentences,  # type: ignore
            max_length=512,
            padding=True,
            truncation=True,
            return_tensors="pt",
        )

        return batch_dict.to(self.model.device)

    def get_embedding_from_output(self, output: ModelOutput, batch_dict: BatchEncoding) -> torch.Tensor:
        return self.average_pool(output.last_hidden_state, batch_dict["attention_mask"])  # type: ignore

    @staticmethod
    def average_pool(last_hidden_states: torch.Tensor, attention_mask: torch.Tensor) -> Tensor:
        last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)
        return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]

    def encode(
        self,
        sentences: list[str],
        *,
        task: Optional[Task] = None,
        batch_size: int = 32,
        encode_type: EncodeTypes = "query",
        **kwargs: Any,  # noqa
    ) -> np.ndarray:
        if self.max_batch_size and batch_size > self.max_batch_size:
            batch_size = self.max_batch_size
        batched_embeddings = []
        if task is not None:  # noqa
            instruction = task_to_instruction(task)
        else:
            instruction = ""
        for batch in tqdm(batched(sentences, batch_size)):
            with torch.inference_mode():
                batch_dict = self.preprocess(batch, instruction=instruction, encode_type=encode_type)
                outputs = self.model(**batch_dict)
                embeddings = self.get_embedding_from_output(outputs, batch_dict)
            batched_embeddings.append(embeddings.detach().cpu())

        return torch.cat(batched_embeddings).to("cpu").detach().numpy()

    def encode_corpus(self, corpus: list[dict[str, str]], **kwargs: Any) -> np.ndarray:
        sep = " "
        if isinstance(corpus, dict):
            sentences = [
                (corpus["title"][i] + sep + corpus["text"][i]).strip() if "title" in corpus else corpus["text"][i].strip()  # type: ignore
                for i in range(len(corpus["text"]))  # type: ignore
            ]
        else:
            sentences = [(doc["title"] + sep + doc["text"]).strip() if "title" in doc else doc["text"].strip() for doc in corpus]
        return self.encode(sentences, encode_type="passage", **kwargs)

    def encode_queries(self, queries: list[str], **kwargs: Any) -> np.ndarray:
        return self.encode(queries, encode_type="query", **kwargs)


class E5Mistral(E5Instruct):
    def __init__(self):
        super().__init__("intfloat/e5-mistral-7b-instruct", max_length=4096, max_batch_size=4, torch_dtype=torch.float16)

    @staticmethod
    def last_token_pool(last_hidden_states: Tensor, attention_mask: Tensor) -> Tensor:
        left_padding = attention_mask[:, -1].sum() == attention_mask.shape[0]
        if left_padding:
            return last_hidden_states[:, -1]
        sequence_lengths = attention_mask.sum(dim=1) - 1
        batch_size = last_hidden_states.shape[0]
        return last_hidden_states[
            torch.arange(batch_size, device=last_hidden_states.device),
            sequence_lengths,
        ]

    def get_embbeding_from_output(self, output: ModelOutput, batch_dict: BatchEncoding) -> torch.Tensor:
        return self.last_token_pool(output.last_hidden_state, batch_dict["attention_mask"])  # type: ignore

    def preprocess(self, sentences: Sequence[str], instruction: str, encode_type: EncodeTypes) -> BatchEncoding:
        if encode_type == "query":
            sentences = [f"Instruction: {instruction}\nQuery: {sentence}" for sentence in sentences]
        batch_dict: BatchEncoding = self.tokenizer(
            sentences,  # type: ignore
            max_length=self.max_length - 1,
            return_attention_mask=False,
            padding=False,
            truncation=True,
        )
        # append eos_token_id to every input_ids
        batch_dict["input_ids"] = [
            [*input_ids, self.tokenizer.eos_token_id]
            for input_ids in batch_dict["input_ids"]  # type: ignore
        ]
        batch_dict = self.tokenizer.pad(batch_dict, padding=True, return_attention_mask=True, return_tensors="pt")

        return batch_dict.to(self.model.device)


@models.register("e5-mistral-7b-instruct")
def create_multilingual_e5_mistral_7b_instruct() -> SebModel:
    hf_name = "intfloat/e5-mistral-7b-instruct"
    meta = ModelMeta(
        name=hf_name.split("/")[-1],
        huggingface_name=hf_name,
        reference=f"https://huggingface.co/{hf_name}",
        languages=[],
        open_source=True,
        embedding_size=4096,
        architecture="Mistral",
        release_date=date(2023, 12, 20),
    )
    return SebModel(
        encoder=LazyLoadEncoder(E5Mistral),
        meta=meta,
    )


@models.register("multilingual-e5-large-instruct")
def create_multilingual_e5_large_instruct() -> SebModel:
    hf_name = "intfloat/multilingual-e5-large-instruct"
    meta = ModelMeta(
        name=hf_name.split("/")[-1],
        huggingface_name=hf_name,
        reference=f"https://huggingface.co/{hf_name}",
        languages=[],
        open_source=True,
        embedding_size=1024,
        architecture="XLM-R",
        release_date=date(2023, 12, 20),
    )
    partial_model = partial(E5Instruct, model_name=hf_name, max_length=512)
    return SebModel(
        encoder=LazyLoadEncoder(partial_model),
        meta=meta,
    )



import random
from collections.abc import Iterable
from itertools import islice
from typing import Any, TypeVar

import datasets
from mteb.abstasks import AbsTaskClustering

from seb.interfaces.mteb_task import MTEBTask
from seb.interfaces.task import Task
from seb.registries import tasks

T = TypeVar("T")


def batched(iterable: Iterable[T], n: int) -> Iterable[tuple[T, ...]]:
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


class HistoricalDanishClustering(AbsTaskClustering):
    @property
    def description(self) -> dict[str, Any]:
        return {
            "name": "HistoricalDanishClustering",
            "hf_hub_name": "kardosdrur/historical-danish-clustering",
            "description": "Subset of the MeMo corpus for clustering historical texts.",
            "reference": "https://huggingface.co/datasets/MiMe-MeMo/Corpus-v1.1",
            "type": "Clustering",
            "category": "p2p",
            "eval_splits": ["train"],
            "eval_langs": ["da"],
            "main_score": "v_measure",
            "revision": "3312e730dfa51f9e411dd4f178e242d00063b363",
        }

    def load_data(self, **kwargs: dict):  # noqa: ARG002
        """
        Load dataset from HuggingFace hub
        """
        if self.data_loaded:
            return

        self.dataset: datasets.DatasetDict = datasets.load_dataset(
            self.description["hf_hub_name"],
            revision=self.description.get("revision"),
        )  # type: ignore

        self.dataset_transform()
        self.data_loaded = True

    def dataset_transform(self):
        splits = self.description["eval_splits"]

        documents: list = []
        labels: list = []
        label_col = "full_title"

        ds = {}
        for split in splits:
            ds_split = self.dataset[split]

            _label = ds_split[label_col]
            documents.extend(ds_split["text"])
            labels.extend(_label)

            assert len(documents) == len(labels)

            rng = random.Random(42)  # local only seed
            pairs = list(zip(documents, labels))
            rng.shuffle(pairs)
            documents, labels = (list(collection) for collection in zip(*pairs))
            ds[split] = datasets.Dataset.from_dict(
                {
                    "sentences": [documents],
                    "labels": [labels],
                }
            )
        self.dataset = datasets.DatasetDict(ds)


@tasks.register("HistoricalDanishClustering")
def create_historical() -> Task:
    task = MTEBTask(HistoricalDanishClustering())
    task.name = "HistoricalDanishClustering"
    task.domain = ["poetry", "fiction"]
    task.task_subtypes = ["Thematic Clustering"]
    return task



from typing import Union

import numpy as np
import torch


def normalize_to_ndarray(embeddings: Union[torch.Tensor, np.ndarray, list[np.ndarray], list[torch.Tensor]]) -> np.ndarray:
    if isinstance(embeddings, list):
        if isinstance(embeddings[0], torch.Tensor):
            return torch.cat(embeddings).detach().cpu().numpy()  # type: ignore
        return np.concatenate(embeddings)
    if isinstance(embeddings, torch.Tensor):
        return embeddings.detach().cpu().float().numpy()
    return embeddings



from typing import Any

from seb.interfaces.mteb_task import MTEBTask
from seb.interfaces.task import Task
from seb.registries import tasks


@tasks.register("SweReC")
def create_swerec() -> Task:
    from mteb import SweRecClassification

    class SweRecClassificationCustom(SweRecClassification):
        @property
        def description(self) -> dict[str, Any]:
            return {
                "name": "SweRecClassification",
                "hf_hub_name": "mteb/swerec_classification",
                "description": "A Swedish dataset for sentiment classification on review",
                "reference": "https://aclanthology.org/2023.nodalida-1.20/",
                "type": "Classification",
                "category": "s2s",
                "eval_splits": ["test"],
                "eval_langs": ["sv"],
                "main_score": "accuracy",
                "n_experiments": 10,
                "samples_per_label": 16,
            }

    task = MTEBTask(SweRecClassificationCustom())
    task.name = "SweReC"
    task.domain = ["reviews"]
    task.task_subtypes = ["Sentiment Classification"]
    return task


@tasks.register("DaLAJ")
def create_dalaj() -> Task:
    from mteb import DalajClassification

    task = MTEBTask(DalajClassification())
    task.name = "DaLAJ"
    task.domain = ["fiction", "non-fiction"]
    task.task_subtypes = ["Linguistic Acceptability"]
    return task


@tasks.register("SweFAQ")
def create_swefaq() -> Task:
    from seb.mteb_tasks import SweFaqRetrieval

    task = MTEBTask(SweFaqRetrieval())
    task.name = "SweFAQ"
    task.version = "0.0.1"
    task.domain = ["non-fiction", "web"]
    task._text_columns = ["question", "candidate_answer", "correct_answer"]
    task.task_subtypes = ["Question-answering"]
    return task


@tasks.register("SwednRetrieval")
def create_swedn_retrieval() -> Task:
    from seb.mteb_tasks import SwednRetrieval

    task = MTEBTask(SwednRetrieval())
    task.name = "SwednRetrieval"
    task.version = "0.0.1"
    task.domain = ["non-fiction", "news"]
    task.task_subtypes = ["Article Retrieval"]
    return task


@tasks.register("SwednClustering")
def create_swedn_clustering() -> Task:
    from seb.mteb_tasks import SwednClustering

    task = MTEBTask(SwednClustering())
    task.name = "SwednClustering"
    task.version = "0.0.1"
    task.domain = ["non-fiction", "news"]
    task._text_columns = ["sentences"]
    task.task_subtypes = ["Thematic Clustering"]
    return task



from __future__ import annotations

from datetime import date
from functools import partial
from typing import Any

import numpy as np
import torch
from sentence_transformers import SentenceTransformer

from seb.interfaces.model import LazyLoadEncoder, ModelMeta, SebModel
from seb.registries import models


class BGEWrapper:
    """following the hf model card documentation."""

    def __init__(self, model_name: str, **kwargs: Any):  # noqa: ARG002
        self.model_name = model_name
        self.mdl = SentenceTransformer(model_name)
        self.sep = " "

    def to(self, device: torch.device) -> None:
        self.mdl.to(device)

    @staticmethod
    def reduce_max_len(sentences: list[str], max_len: int = 10_000) -> list[str]:
        _sentences = []
        for sent in sentences:
            _sentences.append(sent[:max_len])
        return _sentences

    def encode(  # type: ignore
        self,
        sentences: list[str],
        *,
        batch_size: int = 32,
        **kwargs: Any,
    ) -> np.ndarray:
        if "task" in kwargs:
            kwargs.pop("task")

        return np.asarray(self.mdl.encode(self.reduce_max_len(sentences), batch_size=batch_size, **kwargs))

    def encode_queries(self, queries: list[str], batch_size: int = 32, **kwargs: Any) -> np.ndarray:
        if "task" in kwargs:
            kwargs.pop("task")
        sentences = ["Represent this sentence for searching relevant passages: " + sentence for sentence in queries]

        if "convert_to_tensor" in kwargs:
            kwargs.pop("convert_to_tensor")

        emb = self.mdl.encode(self.reduce_max_len(sentences), batch_size=batch_size, normalize_embeddings=True, convert_to_numpy=True, **kwargs)
        return emb.astype("float16")  # type: ignore

    def encode_corpus(
        self,
        corpus: list[dict[str, str]] | dict[str, list[str]],
        batch_size: int = 32,
        **kwargs: Any,
    ) -> np.ndarray:
        if "task" in kwargs:
            kwargs.pop("task")
        if isinstance(corpus, dict):
            sentences = [
                (corpus["title"][i] + self.sep + corpus["text"][i]).strip() if "title" in corpus else corpus["text"][i].strip()  # type: ignore
                for i in range(len(corpus["text"]))  # type: ignore
            ]
        else:
            sentences = [(doc["title"] + self.sep + doc["text"]).strip() if "title" in doc else doc["text"].strip() for doc in corpus]
        if "convert_to_tensor" in kwargs:
            kwargs.pop("convert_to_tensor")

        emb = self.mdl.encode(self.reduce_max_len(sentences), batch_size=batch_size, normalize_embeddings=True, convert_to_numpy=True, **kwargs)
        return emb.astype("float16")  # type: ignore


@models.register("bge-m3")
def create_bge_m3() -> SebModel:
    hf_name = "BAAI/bge-m3"
    meta = ModelMeta(
        name="bge-m3",
        huggingface_name=hf_name,
        reference=f"https://huggingface.co/{hf_name}",
        languages=[],
        open_source=True,
        embedding_size=1024,
        architecture="XLM-R",
        release_date=date(2024, 5, 28),
    )
    return SebModel(
        encoder=LazyLoadEncoder(partial(BGEWrapper, model_name=hf_name)),  # type: ignore
        meta=meta,
    )


if __name__ == "__main__":
    model = create_bge_m3()
    test = model.encoder.encode(["Hello world", "test"])
    assert test.shape == (2, 1024)



from typing import Any

from seb.interfaces.mteb_task import MTEBTask
from seb.interfaces.task import Task
from seb.mteb_tasks import NorwegianCourtsBitextMining, NorwegianParliamentClassification
from seb.registries import tasks


@tasks.register("NoReC")
def create_norec() -> Task:
    from mteb import NoRecClassification

    class NoRecClassificationCustom(NoRecClassification):
        @property
        def description(self) -> dict[str, Any]:
            return {
                "name": "NoRecClassification",
                "hf_hub_name": "mteb/norec_classification",
                "description": "A Norwegian dataset for sentiment classification on review",
                "reference": "https://aclanthology.org/L18-1661/",
                "type": "Classification",
                "category": "s2s",
                "eval_splits": ["test"],
                "eval_langs": ["nb"],
                "main_score": "accuracy",
                "n_experiments": 10,
                "samples_per_label": 16,
            }

    task = MTEBTask(NoRecClassificationCustom())
    task.name = "NoReC"
    task.domain = ["reviews"]
    task.task_subtypes = ["Sentiment Classification"]
    return task


@tasks.register("Norwegian parliament")
def create_norwegian_parliament() -> Task:
    task = MTEBTask(NorwegianParliamentClassification())
    task.name = "Norwegian parliament"
    task.domain = ["spoken"]
    task.task_subtypes = ["Political Classification"]
    return task


@tasks.register("Norwegian courts")
def create_norwegian_courts() -> Task:
    task = MTEBTask(NorwegianCourtsBitextMining())
    task.name = "Norwegian courts"
    task.domain = ["legal", "non-fiction"]
    task._text_columns = ["sentence1", "sentence2"]
    task.task_subtypes = ["Written form Pairing"]
    return task


@tasks.register("VG Clustering")
def create_vg_clustering() -> Task:
    from seb.mteb_tasks import VGClustering

    task = MTEBTask(VGClustering())
    task.name = "VG Clustering"
    task.version = "0.0.1"
    task.domain = ["non-fiction", "news"]
    task._text_columns = ["sentences"]
    task.task_subtypes = ["Thematic Clustering"]
    return task


@tasks.register("NorQuad")
def create_norquad() -> Task:
    from seb.mteb_tasks import NorQuadRetrieval

    task = MTEBTask(NorQuadRetrieval())
    task.name = "NorQuad"
    task.version = "0.0.1"
    task.domain = ["non-fiction", "wiki"]
    task.task_subtypes = ["Question-answering"]
    return task



"""
The openai embedding api's evaluated on the SEB benchmark.
"""

from __future__ import annotations

import logging
import time
from datetime import date
from functools import partial, wraps
from typing import Any, Literal

import numpy as np

from seb.interfaces.model import Encoder, LazyLoadEncoder, ModelMeta, SebModel
from seb.registries import models

logger = logging.getLogger(__name__)


def token_limit(max_tpm: int, interval: int = 60):  # noqa
    limit_interval_start_ts = time.time()
    used_tokens = 0

    def decorator(func):  # noqa
        @wraps(func)
        def wrapper(*args, **kwargs):  # noqa
            nonlocal limit_interval_start_ts, used_tokens

            result = func(*args, **kwargs)
            used_tokens += result.total_tokens

            current_time = time.time()
            if current_time - limit_interval_start_ts > interval:
                limit_interval_start_ts = current_time
                used_tokens = 0

            if used_tokens > max_tpm:
                time.sleep(interval - (current_time - limit_interval_start_ts))
                used_tokens = 0
            return result

        return wrapper

    return decorator


def rate_limit(max_rpm: int, interval: int = 60):  # noqa
    request_interval = interval / max_rpm
    previous_call_ts: float | None = None

    def decorator(func):  # noqa
        @wraps(func)
        def wrapper(*args, **kwargs):  # noqa
            current_time = time.time()
            nonlocal previous_call_ts
            if previous_call_ts is not None and current_time - previous_call_ts < request_interval:
                time.sleep(request_interval - (current_time - previous_call_ts))

            result = func(*args, **kwargs)
            previous_call_ts = time.time()
            return result

        return wrapper

    return decorator


class VoyageWrapper(Encoder):
    def __init__(
        self,
        model_name: str,
        max_retries: int = 5,
        max_rpm: int = 300,
        max_tpm: int = 120_000,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        try:
            import voyageai  # type: ignore
        except ImportError as e:
            raise ImportError("Please install voyageai to use this model using `pip install 'seb[voyageai]'`") from e

        self._client = voyageai.Client(max_retries=max_retries)
        self._embed_func = rate_limit(max_rpm)(token_limit(max_tpm)(self._client.embed))
        self._model_name = model_name
        self._max_tpm = max_tpm
        self.sep = " "

    def encode(self, sentences: list[str], *, batch_size: int = 32, **kwargs: Any) -> np.ndarray:  # noqa: ARG002
        return self._batched_encode(sentences, batch_size, "document")

    def encode_queries(self, queries: list[str], *, batch_size: int = 32, **kwargs: Any) -> np.ndarray:  # noqa: ARG002
        return self._batched_encode(queries, batch_size, "query")

    def encode_corpus(
        self,
        corpus: list[dict[str, str]] | dict[str, list[str]],
        *,
        batch_size: int = 32,
        **kwargs: Any,  # noqa: ARG002
    ) -> np.ndarray:
        if isinstance(corpus, dict):
            sentences = [
                (corpus["title"][i] + self.sep + corpus["text"][i]).strip() if "title" in corpus else corpus["text"][i].strip()  # type: ignore
                for i in range(len(corpus["text"]))  # type: ignore
            ]
        else:
            sentences = [(doc["title"] + self.sep + doc["text"]).strip() if "title" in doc else doc["text"].strip() for doc in corpus]

        return self._batched_encode(sentences, batch_size, "document")

    def _batched_encode(
        self,
        sentences: list[str],
        batch_size: int,
        input_type: Literal["query", "document"],
    ) -> np.ndarray:
        embeddings, index = [], 0

        while index <= len(sentences) - 1:
            batch, batch_tokens = [], 0
            while index < len(sentences) and len(batch) < batch_size and batch_tokens < self._max_tpm:
                n_tokens = len(self._client.tokenize([sentences[index]], model=self._model_name)[0])
                if batch_tokens + n_tokens > self._max_tpm:
                    break
                batch_tokens += n_tokens
                batch.append(sentences[index])
                index += 1

            embeddings.extend(
                self._embed_func(
                    texts=batch,
                    model=self._model_name,
                    input_type=input_type,
                    truncation=True,
                ).embeddings
            )

        return np.array(embeddings)


@models.register("voyage-multilingual-2")
def create_voyage_multilingual_2() -> SebModel:
    api_name = "voyage-multilingual-2"
    meta = ModelMeta(
        name=api_name,
        huggingface_name=None,
        reference="https://blog.voyageai.com/2024/06/10/voyage-multilingual-2-multilingual-embedding-model/",
        languages=[],
        open_source=False,
        embedding_size=1024,
        architecture="API",
        release_date=date(2024, 6, 10),
    )
    return SebModel(
        encoder=LazyLoadEncoder(partial(VoyageWrapper, model_name=api_name)),  # type: ignore
        meta=meta,
    )


if __name__ == "__main__":
    model = create_voyage_multilingual_2()
    test = model.encoder.encode(["Hello world"])



from datetime import date
from functools import partial
from typing import Any, Literal, Optional

import numpy as np
from sentence_transformers import SentenceTransformer

from seb.interfaces.model import LazyLoadEncoder, ModelMeta, SebModel
from seb.interfaces.task import Task
from seb.registries import models

from .normalize_to_ndarray import normalize_to_ndarray
from .sentence_transformer_models import silence_warnings_from_sentence_transformers, wrap_sentence_transformer


class ArcticEncoderWithTaskEncode(SentenceTransformer):
    """
    A sentence transformer wrapper that allows for encoding with a task.
    """

    def encode(  # type: ignore
        self,
        sentences: list[str],
        *,
        batch_size: int = 32,
        task: Optional[Task] = None,
        encode_type: Literal["query", "passage"] = "passage",
        **kwargs: Any,
    ) -> np.ndarray:
        task_prompt = None
        if task is not None and task.task_type in ["Retrieval"] and encode_type == "query":
            task_prompt = "query"

        if task_prompt is None:
            emb = super().encode(sentences, batch_size=batch_size, **kwargs)
        else:
            emb = super().encode(sentences, batch_size=batch_size, task=task_prompt, prompt_name=task_prompt, **kwargs)
        return normalize_to_ndarray(emb)

    def encode_corpus(self, corpus: list[dict[str, str]], **kwargs: Any) -> np.ndarray:
        sep = " "
        if isinstance(corpus, dict):
            sentences = [
                (corpus["title"][i] + sep + corpus["text"][i]).strip() if "title" in corpus else corpus["text"][i].strip()  # type: ignore
                for i in range(len(corpus["text"]))  # type: ignore
            ]
        else:
            sentences = [(doc["title"] + sep + doc["text"]).strip() if "title" in doc else doc["text"].strip() for doc in corpus]
        return self.encode(sentences, encode_type="passage", **kwargs)

    def encode_queries(self, queries: list[str], **kwargs: Any) -> np.ndarray:
        return self.encode(queries, encode_type="query", **kwargs)


def wrap_arctic_sentence_transformer(model_name: str, max_seq_length: Optional[int] = None, **kwargs: Any) -> ArcticEncoderWithTaskEncode:
    silence_warnings_from_sentence_transformers()
    mdl = ArcticEncoderWithTaskEncode(model_name, **kwargs)
    if max_seq_length is not None:
        mdl.max_seq_length = max_seq_length
    return mdl


@models.register("snowflake-arctic-embed-l-v2.0")
def create_artic_embed_l_v2() -> SebModel:
    hf_name = "Snowflake/snowflake-arctic-embed-l-v2.0"
    meta = ModelMeta(
        name=hf_name.split("/")[-1],
        huggingface_name=hf_name,
        reference=f"https://huggingface.co/{hf_name}",
        languages=[],
        open_source=True,
        embedding_size=1024,
        architecture="XLM-R",
        release_date=date(2024, 11, 8),
    )
    return SebModel(
        encoder=LazyLoadEncoder(partial(wrap_arctic_sentence_transformer, model_name=hf_name, trust_remote_code=True)),  # type: ignore
        meta=meta,
    )


@models.register("snowflake-arctic-embed-m-v2.0")
def create_artic_embed_m_v2() -> SebModel:
    hf_name = "Snowflake/snowflake-arctic-embed-m-v2.0"
    meta = ModelMeta(
        name=hf_name.split("/")[-1],
        huggingface_name=hf_name,
        reference=f"https://huggingface.co/{hf_name}",
        languages=[],
        open_source=True,
        embedding_size=768,
        architecture="GTE",
        release_date=date(2024, 11, 8),
    )
    return SebModel(
        encoder=LazyLoadEncoder(partial(wrap_arctic_sentence_transformer, model_name=hf_name, trust_remote_code=True)),  # type: ignore
        meta=meta,
    )



import random
from collections.abc import Iterable
from itertools import islice
from typing import Any, TypeVar

import datasets
from mteb.abstasks import AbsTaskClustering

T = TypeVar("T")


def batched(iterable: Iterable[T], n: int) -> Iterable[tuple[T, ...]]:
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


class SwednClustering(AbsTaskClustering):
    @property
    def description(self) -> dict[str, Any]:
        return {
            "name": "SwednClustering",
            "hf_hub_name": "sbx/superlim-2",
            "description": "The SWE-DN corpus is based on 1,963,576 news articles from the Swedish newspaper "
            + "Dagens Nyheter (DN) during the years 2000--2020. The articles are filtered to resemble the CNN/DailyMail"
            + " dataset both regarding textual structure. This dataset uses the category labels as clusters.",
            "reference": "https://spraakbanken.gu.se/en/resources/swedn",
            "type": "Clustering",
            "category": "p2p",
            "eval_splits": ["all"],
            "eval_langs": ["sv"],
            "main_score": "v_measure",
            "revision": "ef1661775d746e0844b299164773db733bdc0bf6",
        }

    def load_data(self, **kwargs: dict):  # noqa: ARG002
        """
        Load dataset from HuggingFace hub
        """
        if self.data_loaded:
            return

        self.dataset: datasets.DatasetDict = datasets.load_dataset(
            self.description["hf_hub_name"],
            "swedn",  # chose the relevant subset
            revision=self.description.get("revision"),
        )  # type: ignore

        self.dataset_transform()
        self.data_loaded = True

    def dataset_transform(self):
        """
        The article_category clusters differ between the splits (with the test set only having 1 cluster). Therefore we combine it all into one
        cluster.
        """
        splits = ["train", "validation"]
        # performance of sample models with test set: 8.74, 2.43 -removing test-> 11.26, 4.27
        # this is due to the test set only having 1 cluster which is "other"

        headlines = []
        summaries = []
        articles = []
        labels = []
        label_col = "article_category"

        for split in splits:
            ds_split = self.dataset[split]
            headlines.extend(ds_split["headline"])
            labels.extend(ds_split[label_col])

            summaries.extend(ds_split["summary"])
            labels.extend(ds_split[label_col])

            articles.extend(ds_split["article"])
            labels.extend(ds_split[label_col])

        rng = random.Random(42)  # local only seed

        clusters_text = []
        clusters_labels = []
        doc_types = [summaries, articles]
        # Note that headlines is excluded:
        # Scores of sample models with headlines: 11.26, 4.27 -removing headlines-> 16.43, 4.31
        # as headlines are soo short it is hard to meaningfully cluster them even for humans.
        for text in doc_types:
            pairs = list(zip(text, labels))
            rng.shuffle(pairs)
            # reduce size of dataset to not have too large datasets in the clustering task
            pairs_batched = list(batched(pairs, 512))
            texts1, labels2 = list(zip(*pairs_batched[0]))
            texts2, labels2 = list(zip(*pairs_batched[1]))

            clusters_text.extend([texts1, texts2])
            clusters_labels.extend([labels2, labels2])
        ds = datasets.Dataset.from_dict({"sentences": clusters_text, "labels": clusters_labels})
        self.dataset = datasets.DatasetDict({"all": ds})



import random
from collections.abc import Iterable
from itertools import islice
from typing import Any, TypeVar

import datasets
from mteb.abstasks import AbsTaskClustering

T = TypeVar("T")


def batched(iterable: Iterable[T], n: int) -> Iterable[tuple[T, ...]]:
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


class VGClustering(AbsTaskClustering):
    @property
    def description(self) -> dict[str, Any]:
        return {
            "name": "VGClustering",
            "hf_hub_name": "navjordj/VG_summarization",
            "description": "Articles and their classes (e.g. sports) from VG news articles extracted from Norsk Aviskorpus.",
            "reference": "https://huggingface.co/datasets/navjordj/VG_summarization",
            "type": "Clustering",
            "category": "p2p",
            "eval_splits": ["test"],
            "eval_langs": ["nb"],
            "main_score": "v_measure",
            "revision": "d4c5a8ba10ae71224752c727094ac4c46947fa29",
        }

    def load_data(self, **kwargs: dict):  # noqa: ARG002
        """
        Load dataset from HuggingFace hub
        """
        if self.data_loaded:
            return

        self.dataset: datasets.DatasetDict = datasets.load_dataset(
            self.description["hf_hub_name"],
            revision=self.description.get("revision"),
        )  # type: ignore

        self.dataset_transform()
        self.data_loaded = True

    def dataset_transform(self):
        splits = self.description["eval_splits"]

        documents: list = []
        labels: list = []
        label_col = "classes"

        ds = {}
        for split in splits:
            ds_split = self.dataset[split]

            _label = self.normalize_labels(ds_split[label_col])
            documents.extend(ds_split["title"])
            labels.extend(_label)

            documents.extend(ds_split["ingress"])
            labels.extend(_label)

            documents.extend(ds_split["article"])
            labels.extend(_label)

            assert len(documents) == len(labels)

            rng = random.Random(1111)  # local only seed
            # resampling changes scores from 12.68, 11.30, 12.65 (sample model)
            pairs = list(zip(documents, labels))
            rng.shuffle(pairs)
            documents, labels = (list(collection) for collection in zip(*pairs))

            # reduce size of dataset to not have too large datasets in the clustering task
            documents_batched = list(batched(documents, 512))[:4]
            labels_batched = list(batched(labels, 512))[:4]
            # See:
            # https://github.com/KennethEnevoldsen/scandinavian-embedding-benchmark/pull/96
            # for a discussion on sizes

            ds[split] = datasets.Dataset.from_dict(
                {
                    "sentences": documents_batched,
                    "labels": labels_batched,
                }
            )

        self.dataset = datasets.DatasetDict(ds)

    @staticmethod
    def normalize_labels(labels: list[str]) -> list[str]:
        # Agreed on and debated in: https://github.com/KennethEnevoldsen/scandinavian-embedding-benchmark/issues/83
        return [label.split(",")[0] for label in labels]
